#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

/////////////////////////////////////////////////////////////////////////////////////////////////
// File:					dmus_demo.cpp
// 
// Related Header File:		None
//
// Original Author:			Michael "Aldacron" Parker
//
// Creation Date:			10/8/2000
//
// Purpose:					Demonstrates using DirectMusic to load and play MIDI files
//							in a Windows application. Note that there is no main loop.
//							To follow program flow, take a look at the Windows message
//							handlers in the WinProc function below. This file accompanies
//							the article "DirectMusic Symphony in C++, Part One". Feel
//							free to utilize this code in any method you desire.
//		
//							Thanks to Andre' LaMothe for 2 things - 1) for writing
//							the books that taught me this stuff and 2) for
//							supplying the free MIDIs on his Tricks of the Windows
//							Game Programming Gurus CD. Battle.mid comes off of that
//							CD.
////////////////////////////////////////////////////////////////////////////////////////////////
// we must either define INITGUID or link to dxguid.lib
// I prefer to define it like so:
#define INITGUID
#include <windows.h>
#include <string> // for wcscpy
// required DirectMusic includes
#include <dmksctrl.h>
#include <dmusici.h>
#include <dmusicc.h>
#include <dmusicf.h>

#define APPNAME "DirectMusic Demo 1"
// macro to save typing
#define MSGBOX(output)  MessageBox(NULL,output,"DMusic Demo",MB_OK);
// macro to convert ascii strings to wide-character format. This same
// macro can be found in the DirectMusic documentation in the
// tutorial on loading a MIDI file, as well as in the book
// Tricks of the Windows Game Programming Gurus by Andre' LaMothe.
#define MULTI_TO_WIDE( x,y )  MultiByteToWideChar( CP_ACP,MB_PRECOMPOSED, y,-1,x,_MAX_PATH);

HWND main_hwnd;									// handle to the applcation's window
IDirectMusicPerformance *performance = NULL;	// the performance object controls manipulation of
												// the segments
IDirectMusicSegment *segment;					// the segment object is where the MIDI data is
												// stored for playback
IDirectMusicLoader *loader;						// the loader bject can load many types of
												// DMusic related files


////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			InitDirectMusic
//
// Last Modified:			10/8/2000
//
// Function:				Demonstrates initializing COM and DirectMusic
//
// Inputs:					None
//
// Ouputs:					None
//
// Returns:					Nothing
///////////////////////////////////////////////////////////////////////////////////////////////
bool InitDirectMusic (void)
{
	HRESULT hr;
	// first, make sure COM is initialized. This is nessecary because
	// DirectMusic objects must be created through COM. Normally, you
	// would place this in a WM_CREATE handler, or in your game's
	// start-up code. It's placed here for demonstration purposes
	CoInitialize (NULL);
	// now that COM is initialized, we can get busy. We'll start by creating
	// the performance object via CoCreateInstance. I hope the comments
	// don't damage the readability too much...
	if (FAILED(CoCreateInstance(
			(REFCLSID)CLSID_DirectMusicPerformance, // class identifier
													// of the object to
													// create
			NULL, // always NULL
			CLSCTX_INPROC, // context for running executable code (just
						   // always use CLSTCX_INPROC) 
			(REFIID)IID_IDirectMusicPerformance, // reference to the
												 // interface identifier
			(LPVOID *)&performance))) // address of the ouput variable that
									  // recieves the interface pointter
	{
		MSGBOX ("Failed to create the performance object");
		return false;
	}
	// now that the performance object has been successfully created,
	// we need to initialize it
	hr = performance->Init (NULL, // address to a DirectMusic object, if
								  // it exists. If NULL, the object will
								  // be created internally
					   NULL, // address of a DirectSound object. If you use
							// DirectSound in conjuction with DMusic, you
							// should ALWAYS pass the DSound pointer. We
							// aren't using DSound, so we can pass NULL
							// and the object will be created internally.
					   main_hwnd); // application's window handle
	// check the HRESULT returned by the function. Normally, I convert
	// HRESULT value to a string and save it to a log file so that I
	// know the exact error. For our purposes, I'll just pop-up a
	// message box when a function fails
	if (FAILED(hr))
	{
		MSGBOX("Failed to initialize performance object");
		return false;
	}
	// now we need to choose a port. As stated in the article, you could
	// enumerate all available ports and let the user choose one, but it's
	// much simpler to just use the default Windows synth by passing
	// NULL as the port parameter. Sounds better, too.
	hr = performance->AddPort (NULL);
	if (FAILED(hr))
	{
		MSGBOX("AddPort failed");
		return false;
	}
	// now we'll create the loader object. This will be used to load the
	// midi file for our demo. Again, we need to use CoCreateInstance
	// and pass the appropriate ID parameters
	if (FAILED(CoCreateInstance ((REFCLSID)CLSID_DirectMusicLoader,
								  NULL,
								  CLSCTX_INPROC,
								  (REFIID)IID_IDirectMusicLoader,
								  (LPVOID *)&loader)))
	{
		MSGBOX("Failed to create loader object");
		return false;
	}
	// that's it for initialization. If we made it this far we
	// were successful.
	return true;
}
////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			ShutdownDirectMusic
//
// Last Modified:			10/11/00
//
// Function:				Releases memory used by all of the initialized
//							DirectMusic objects in the program
//
// Inputs:					None
//
// Ouputs:					None
//
// Returns:					Nothing
///////////////////////////////////////////////////////////////////////////////////////////////
void ShutdownDirectMusic (void)
{
	// it's always a good idea to release DirectX objects in the
	// order you created them, so the first thing we'll do is kill
	// the segment. This is done by making a call to 
	// IDIrectMusicSegment::SetParam and passing GUID_Unload as the
	// parameter to set. The inportant parts are the first and
	// last parameters. You will usually (always) pass the
	// performance object as the last parameter, and the reference
	// ID of the segement parameter you wish to set as the first.
	// The other 3 arguments are for advanced DMusic stuff, which
	// we won't get into here. Just always use -1,0,0 for now.
	if (NULL != segment)
	{
		segment->SetParam ((REFGUID)GUID_Unload,
						   -1,
						   0,
						   0,
						   (LPVOID *)performance);
		segment->Release ();
		segment = NULL;
	}
	// next up is the loader object. Nothing special here. Just call Release.
	if (NULL != loader)
	{
		loader->Release ();
		loader = NULL;
	}
	// and finally the performance. This requires a call to
	// IDirectMusicPerformance::CloseDown, which takes no
	// arguments, before calling Release
	if (NULL != performance)
	{
		performance->CloseDown ();
		performance->Release ();
		performance = NULL;
	}
	// and now we need to call CoUninitialize in order to shutdown
	// COM. Normally, this whould go in a WM_CLOSE or WM_DESTROY handler,
	// or in your game's main shutdown function, but is here for
	// demonstration purposes
	CoUninitialize ();
}
////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			LoadMIDI
//
// Last Modified:			10/11/00
//
// Function:				Demonstrates how to use the IDirectMusicLoader interface
//							to load MIDI files from disk
//
// Inputs:					filename = name of the MIDI file to load
//
// Ouputs:					None
//
// Returns:					Nothing
///////////////////////////////////////////////////////////////////////////////////////////////
bool LoadMIDI (char *filename)
{
	// this is probably the most complicated function we'll deal with in this app.
	// I've tried to comment everything in detail, since I didn't put any details
	// in the article.
	// the DMUS_OBJECTDESC structure works like a surface desc in DDraw, or a buffer
	// desc in DSound. It is used by the loader object to identify the object it is
	// supposed to load
	DMUS_OBJECTDESC obj_desc;
	HRESULT hr;		// placeholder for function return values
	// for reference, _MAX_PATH is a Win32 API defined constant that represents
	// the maximum length of a path name in Windows
	char directory[_MAX_PATH];		// placeholder for the name of the directory in which the
									// loader should search for the file
	WCHAR w_directory[_MAX_PATH];	// wide-character (unicode) version of the directory name.
									// This loader will use this variable instead of the
									// regular ASCII string above
	WCHAR w_filename[_MAX_PATH];	// unicode string containing the name of the file the
									// loader should load
	// first thing we want to do is get the current working directory. For this, we can
	// use the Win32 API function GetCurrentDirectory (or the equivalent C/C++ functions
	// if you prefer). The function returns 0 on failure.
	if (0 == GetCurrentDirectory (_MAX_PATH,directory))
	{
		MSGBOX("LoadMIDI: GetCurrentDirectory failed");
		return false;
	}
	// now we need to convert the ASCII directory name to wide-character format.
	// We can do this with our handy macro, MULTI_TO_WIDE
	MULTI_TO_WIDE(w_directory,directory);
	// now we need to tell the loader to start its search in the current directory with the
	// fuinction IDirectMusicLoader::SetSearchDirectory. The first parameter tells what type
	// of object to load. For our purposes, we'll use GUID_DirectMusicAllTypes, which means
	// to load any supported objects. I just use it by default. The second
	// parameter is the directory name we obtained above. The 3rd parameter, according
	// to the docs, tells the loader if it should clear information about previous objects
	// in order to avoid loading objects of the same name from the directory that
	// was set previously. I just always use FALSE, but you could set it to TRUE if you
	// like. Note that this parameter is a Win32 BOOL and not a C++ bool. You could even call
	// this function, in conjunction with the above steps, in the InitDirectMusic function
	// if you wanted to. Also note that if your MIDI file is in a subdirectory off of the
	// current directory, the loader will still find it. This function just tells it
	// where the root of its search should be - it does not need to be the current directory.
	// It's just much easier to do it this way, especially if you don't know the full path
	// to the dir where your MIDI is stored.
	hr = loader->SetSearchDirectory ((REFGUID)GUID_DirectMusicAllTypes, w_directory, FALSE);
	if (FAILED(hr))
	{
		MSGBOX("LoadMIDI: SetSearchDirectory failed");
		return false;
	}
	// now we need to set up the object description. First, clear the memory and set
	// the dwSize parameter (like you do with all DirectX structures).
	ZeroMemory (&obj_desc, sizeof (obj_desc));
	obj_desc.dwSize = sizeof (obj_desc);
	// first, change the ascii filename to wide-character format
	MULTI_TO_WIDE(w_filename,filename);
	// the loader needs to know the class id of the object it is going to load.
	// Since we want to load a MIDI file, which will be stored in an
	// IDirectMusicSegment object, we need to pass the class id of
	// CLSID_DirectMusicSegment.
	obj_desc.guidClass = CLSID_DirectMusicSegment;
	// the loader object also needs to know the filename. We'll set this up
	// by using the function wcscpy from string.h, since we're using
	// wide-character strings
	wcscpy (obj_desc.wszFileName,w_filename);
	// now we also need to tell the loader which members of the object
	// description we are using, in our case the guidClass and wszFilename
	// members
	obj_desc.dwValidData = DMUS_OBJ_CLASS | DMUS_OBJ_FILENAME;
	// now that everything is set up, we can call IDirectMusicLoader::GetObject
	// to load the file into the segment. We need to pass the object description,
	// the reference ID of the type of object to create (in this case a 
	// DirectMusicSegment) and the address where a pointer to the object will
	// be stored.
	hr = loader->GetObject (&obj_desc,
							(REFIID)IID_IDirectMusicSegment,
							(LPVOID *) &segment);
	if (FAILED(hr))
	{
		MSGBOX("LoadMIDI: Get object failed");
		return FALSE;
	}
	// next we need to tell the segment what kind of data it contains. We do this
	// with the IDirectMusicSegment::SetParam function.
	hr = segment->SetParam ((REFGUID)GUID_StandardMIDIFile,
							-1,
							0,
							0,
							(LPVOID)performance);
	if (FAILED(hr))
	{
		MSGBOX("LoadMIDI: SetParam (MIDI file) failed");
		return false;
	}
	// finally, we need to tell the segment to 'download' the instruments
	// required to play this file. This is all handled internally, so all
	// you need to do is to say "Do it!".
	hr = segment->SetParam ((REFGUID)GUID_Download,-1,0,0,(LPVOID)performance);
	if (FAILED(hr))
	{
		MSGBOX("LoadMIDI: Failed to download instruments");
		return false;
	}
	// at this point, the MIDI file is loaded and ready to play!
	return true;
}
////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			PlaySegment
//
// Last Modified:			10/11/00
//
// Function:				Plays a MIDI that has been loaded into a DirectMusicSegment object.
//
// Inputs:					None
//
// Ouputs:					None
//
// Returns:					Nothing
///////////////////////////////////////////////////////////////////////////////////////////////
void PlaySegment (void)
{
	HRESULT hr;
	// okay, there are no parameters to this function since we have a global segment
	// object. All we need to do is tell the performance to play that segment by passing
	// a pointer to it as a parameter. Normally, I don't check for errors here. But for the
	// demo, I'll do it anyway.
	hr = performance->PlaySegment (segment, // segment to play
								   0, // behavior flags - advanced (just use 0)
								   0, // start time - advanced (just use 0)
								   NULL); // segment state - advanced (just set NULL)
	if (FAILED(hr))
	{
		MSGBOX("PlaySegment failed");
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			StopSegment
//
// Last Modified:			10/11/00
//
// Function:				Stops a currently playing segment
//
// Inputs:					None
//
// Ouputs:					None
//
// Returns:					Nothing
///////////////////////////////////////////////////////////////////////////////////////////////
void StopSegment (void)
{
	HRESULT hr;
	hr = performance->Stop (segment, // segment to stop playing
						    NULL,  // associated SegmentState (just use NULL)
						    0, // time to stop playing (just use 0)
						    0); // exact boundary at which to stop (0 = immediately)
	if (FAILED(hr))
	{
		MSGBOX("StopSegment failed");
	}
}
////////////////////////////////////////////////////////////////////////////////////////////////
// Function Name:			IsSegmentPlaying
//
// Last Modified:			10/11/00
//
// Function:				Determines if the segment is currently playing or not.
//
// Inputs:					None
//
// Ouputs:					None
//
// Returns:					true if segment is playing, false if not
///////////////////////////////////////////////////////////////////////////////////////////////
bool IsSegmentPlaying (void)
{
	HRESULT hr;
	// IsPlaying return S_OK if the segment is currently playing
	hr = performance->IsPlaying (segment, // segment to test
								 NULL); // associated SegmentState
	if (hr == S_OK)
		return TRUE;
	return FALSE;
}

// i'm not going to comment all of this Windows stuff - you should already
// know this if you're tackling DirectMusic. I will, however, comment the code that is
// specific to this demo
LRESULT CALLBACK WinProc (HWND hwnd,UINT msg,WPARAM wparam,LPARAM lparam)
{
	switch (msg)
	{
	case WM_CREATE:
		// get DirectMusic up and running
		InitDirectMusic ();
		// load our demo MIDI
		LoadMIDI("Battle.mid");
		// now play it
		PlaySegment ();
		return 0;
	case WM_KEYDOWN:
		switch (wparam)
		{
		case VK_ESCAPE:
			SendMessage (hwnd,WM_CLOSE,0,0);
			break;
		case VK_SPACE:
			// stop the segment if it's playing
			if (IsSegmentPlaying ())
			{
				StopSegment ();
			} else
			{
				PlaySegment ();
			}
			break;
		case VK_RETURN:
			// call the IsSegmentPlaying function to see if the
			// midi has stopped playing or not. If it has, reset
			// the mid_is_playing var and start it again. If it
			// hasn't, just display a message box saying so
			if (IsSegmentPlaying ())
			{
				MSGBOX("Segment is still playing");
			} else
			{
				MSGBOX("Segment has stopped. Will play again.");
				PlaySegment ();
			}
			break;
		default:
			break;
		}
		return 0;

	case WM_DESTROY:
		// this is a good place to kill the DirectMusic objects
		ShutdownDirectMusic ();
		PostQuitMessage (1);
		return 0;
	default:
		break;
	}
	return DefWindowProc (hwnd,msg,wparam,lparam);
}

int WINAPI WinMain (HINSTANCE hinstance,HINSTANCE prev,LPSTR cmdline,int cmdshow)
{
	HWND hwnd;
	MSG msg;
	WNDCLASSEX wc;

	wc.cbSize = sizeof(WNDCLASSEX);
	wc.cbClsExtra = 0;
	wc.cbWndExtra = 0;
	wc.hInstance = hinstance;
	wc.lpfnWndProc = WinProc;
	wc.hbrBackground = (HBRUSH)GetStockObject (BLACK_BRUSH);
	wc.hIcon = LoadIcon (NULL,IDI_APPLICATION);
	wc.hIconSm = LoadIcon (NULL,IDI_APPLICATION);
	wc.hCursor = LoadCursor (NULL,IDC_ARROW);
	wc.lpszClassName = APPNAME;
	wc.lpszMenuName = NULL;
	wc.style = CS_HREDRAW | CS_VREDRAW;

	if (!RegisterClassEx (&wc))
	{
		return 0;
	}
	hwnd = CreateWindowEx (WS_EX_TOPMOST,APPNAME,APPNAME,WS_OVERLAPPEDWINDOW,
		CW_USEDEFAULT,CW_USEDEFAULT,300,300,NULL,NULL,hinstance,NULL);
	if (!hwnd)
	{
		return 0;
	}
	main_hwnd = hwnd;
	ShowWindow (hwnd,cmdshow);
	UpdateWindow (hwnd);

	while (GetMessage (&msg,NULL,0,0))
	{
		TranslateMessage (&msg);
		DispatchMessage (&msg);
	}
	return msg.wParam;
}
	
	
